#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
All X11 specific DSI functions
"""

# built-in modules
from typing import Optional
from logging import getLogger, CRITICAL, Logger
from ctypes.util import find_library
from ctypes import cdll
from ctypes import (
    POINTER,
    byref,
    cast,
    c_char_p,
    c_int,
    c_long,
    c_ubyte,
    c_ulong,
    c_char,
    Structure,
    c_int32,
    c_uint,
    c_void_p,
    c_uint32,
    _SimpleCData,
    Union,
    CFUNCTYPE
)

# local modules
from .base import DSIBase
from .window import WindowBase
from .image import Image
from .buttons import MouseButtons
from .box import Box

# Setup Xlib Structures

# pylint: disable=too-few-public-methods

PLAINMASK = 0x00FFFFFF
ZPIXMAP = 2


class Display(Structure):
    """
    https://tronche.com/gui/x/xlib/display/opening.html#Display\n
    /usr/include/X11/Xlib.h: 487
    """


class XImage(Structure):
    """
    https://tronche.com/gui/x/xlib/graphics/images.html#XImage\n
    /usr/include/X11/Xlib.h: 360-394
    """

    _fields_ = [
        ('width', c_int),
        ('height', c_int),
        ('xoffset', c_int),
        ('format', c_int),
        ('data', c_void_p),
        ('byte_order', c_int),
        ('bitmap_unit', c_int),
        ('bitmap_bit_order', c_int),
        ('bitmap_pad', c_int),
        ('depth', c_int),
        ('bytes_per_line', c_int),
        ('bits_per_pixel', c_int),
        ('red_mask', c_ulong),
        ('green_mask', c_ulong),
        ('blue_mask', c_ulong)
    ]


class XWindowAttributes(Structure):
    """
    https://tronche.com/gui/x/xlib/window-information/XGetWindowAttributes.html\n
    /usr/include/X11/Xlib.h: 308-334
    """

    _fields_ = [
        ("x", c_int32),
        ("y", c_int32),
        ("width", c_int32),
        ("height", c_int32),
        ("border_width", c_int32),
        ("depth", c_int32),
        ("visual", c_ulong),
        ("root", c_ulong),
        ("class", c_int32),
        ("bit_gravity", c_int32),
        ("win_gravity", c_int32),
        ("backing_store", c_int32),
        ("backing_planes", c_ulong),
        ("backing_pixel", c_ulong),
        ("save_under", c_int32),
        ("colourmap", c_ulong),
        ("mapinstalled", c_uint32),
        ("map_state", c_uint32),
        ("all_event_masks", c_ulong),
        ("your_event_mask", c_ulong),
        ("do_not_propagate_mask", c_ulong),
        ("override_redirect", c_int32),
        ("screen", c_ulong)
    ]


class XKeyEvent(Structure):
    """
    https://tronche.com/gui/x/xlib/events/keyboard-pointer/keyboard-pointer.html#XKeyEvent\n
    /usr/include/X11/Xlib.h: 557-571
    """

    _fields_ = [
        ('type', c_int),
        ('serial', c_ulong),
        ('send_event', c_int),
        ('display', POINTER(Display)),
        ('window', c_ulong),  # Window / XID
        ('root', c_ulong),  # Window / XID
        ('subwindow', c_ulong),  # Window / XID
        ('time', c_ulong),  # Time
        ('x', c_int),
        ('y', c_int),
        ('x_root', c_int),
        ('y_root', c_int),
        ('state', c_uint),
        ('keycode', c_uint),
        ('same_screen', c_int),
    ]


class XButtonEvent(Structure):
    """
    https://tronche.com/gui/x/xlib/events/keyboard-pointer/keyboard-pointer.html#XButtonEvent\n
    /usr/include/X11/Xlib.h: 575-589
    """

    _fields_ = [
        ('type', c_int),
        ('serial', c_ulong),
        ('send_event', c_int),
        ('display', POINTER(Display)),
        ('window', c_ulong),  # Window (XID)
        ('root', c_ulong),  # Window (XID)
        ('subwindow', c_ulong),  # Window (XID)
        ('time', c_ulong),  # Time
        ('x', c_int),
        ('y', c_int),
        ('x_root', c_int),
        ('y_root', c_int),
        ('state', c_uint),
        ('button', c_uint),
        ('same_screen', c_int),
    ]


class XEvent(Union):
    """
    https://tronche.com/gui/x/xlib/events/structures.html#XEvent\n
    /usr/include/X11/Xlib.h: 973-1009
    """
    _fields_ = [
        ('type', c_int),
        ('xkey', XKeyEvent),
        ('xbutton', XButtonEvent),
        ('pad', c_long*24),
    ]


class XErrorEvent(Structure):
    """
    https://tronche.com/gui/x/xlib/event-handling/protocol-errors/default-handlers.html#XErrorEvent\n
    /usr/include/X11/Xlib.h: 924-932
    """

    def __repr__(self) -> str:
        # pylint: disable-next=line-too-long
        return f"XErrorEvent(type={self.type}, serial={self.serial}, error_code={self.error_code}, request_code={self.request_code}, minor_code={self.minor_code})"

    _fields_ = [
        ("type", c_int),
        ("display", POINTER(Display)),
        ("serial", c_ulong),
        ("error_code", c_ubyte),
        ("request_code", c_ubyte),
        ("minor_code", c_ubyte),
        ("resourceid", c_void_p),
    ]


logger = getLogger(__name__)
logger.setLevel(CRITICAL)


@CFUNCTYPE(c_int, POINTER(Display), POINTER(XErrorEvent))
def error_handler(_, event):
    """
    A C function that handles X11 errors.
    """
    logger.error("%s", event.contents)
    return 0


def get_logger() -> Logger:
    """
    Returns a logger that is responsible for logging XErrorEvents.
    The logger is set to CRITICAL. So it will not log anything.
    To make it log something, set the log level at least to ERROR.
    You can archive this with "logger.setLevel(logging.ERROR)".
    """
    return logger

# Setup Xlib Variables


class Masks:
    """
    https://tronche.com/gui/x/xlib/events/mask.html\n
    /usr/include/X11/X.h: 150-175
    """
    NoEventMask = 0
    KeyPressMask = 1
    KeyReleaseMask = 2
    ButtonPressMask = 4
    ButtonReleaseMask = 8
    EnterWindowMask = 16
    LeaveWindowMask = 32
    PointerMotionMask = 64
    PointerMotionHintMask = 128
    Button1MotionMask = 256
    Button2MotionMask = 512
    Button3MotionMask = 1024
    Button4MotionMask = 2048
    Button5MotionMask = 4096
    ButtonMotionMask = 8192
    KeymapStateMask = 16384
    ExposureMask = 32768
    VisibilityChangeMask = 65536
    StructureNotifyMask = 131072
    ResizeRedirectMask = 262144
    SubstructureNotifyMask = 524288
    SubstructureRedirectMask = 1048576
    FocusChangeMask = 2097152
    PropertyChangeMask = 4194304
    ColormapChangeMask = 8388608
    OwnerGrabButtonMask = 16777216


class EventTypes:
    """
    https://tronche.com/gui/x/xlib/events/types.html\n
    /usr/include/X11/X.h: 181-215
    """
    KeyPress = 2
    KeyRelease = 3
    ButtonPress = 4
    ButtonRelease = 5
    MotionNotify = 6
    EnterNotify = 7
    LeaveNotify = 8
    FocusIn = 9
    FocusOut = 10
    KeymapNotify = 11
    Expose = 12
    GraphicsExpose = 13
    NoExpose = 14
    VisibilityNotify = 15
    CreateNotify = 16
    DestroyNotify = 17
    UnmapNotify = 18
    MapNotify = 19
    MapRequest = 20
    ReparentNotify = 21
    ConfigureNotify = 22
    ConfigureRequest = 23
    GravityNotify = 24
    ResizeRequest = 25
    CirculateNotify = 26
    CirculateRequest = 27
    PropertyNotify = 28
    SelectionClear = 29
    SelectionRequest = 30
    SelectionNotify = 31
    ColormapNotify = 32
    ClientMessage = 33
    MappingNotify = 34
    GenericEvent = 35
    LASTEvent = 36


class KeyMasks:
    """
    https://tronche.com/gui/x/xlib/events/keyboard-pointer/keyboard-pointer.html\n
    /usr/include/X11/X.h: 221-228
    """
    ShiftMask = 1
    LockMask = 2
    ControlMask = 4
    Mod1Mask = 8
    Mod2Mask = 16
    Mod3Mask = 32
    Mod4Mask = 64
    Mod5Mask = 128


class Xlib:
    """
    A class that provides access to Xlib functions.
    """

    def __init__(self):
        # load libX11.so.6
        x11 = find_library("X11")
        if not x11:
            raise FileNotFoundError("X11 library not found!")
        self.xlib = cdll.LoadLibrary(x11)

        self.xlib.XSetErrorHandler(error_handler)

        # Setup Xlib functions
        self.xlib.XGetImage.argtypes = [
            POINTER(Display),
            c_ulong,  # Drawable (XID)
            c_int,
            c_int,
            c_uint,
            c_uint,
            c_ulong,
            c_int,
        ]
        self.xlib.XGetImage.restype = POINTER(XImage)

        self.xlib.XGetWindowAttributes.argtypes = [
            POINTER(Display),
            c_ulong,  # Window (XID)
            POINTER(XWindowAttributes)
        ]
        self.xlib.XOpenDisplay.restype = POINTER(Display)
        self.xlib.XGetWindowProperty.argtypes = [
            POINTER(Display),
            c_ulong,  # Window
            c_ulong,  # Atom
            c_long,
            c_long,
            c_int,
            c_ulong,  # Atom
            POINTER(c_ulong),  # Atom
            POINTER(c_int),
            POINTER(c_ulong),
            POINTER(c_ulong),
            POINTER(POINTER(c_ubyte))
        ]
        self.xlib.XInternAtom.argtypes = [POINTER(Display), c_char_p, c_int]
        self.xlib.XFree.argtypes = [c_void_p]
        self.xlib.XDestroyImage.argtypes = [POINTER(XImage)]
        self.xlib.XWarpPointer.argtypes = [
            POINTER(Display),
            c_ulong,
            c_ulong,
            c_int,
            c_int,
            c_uint,
            c_uint,
            c_int,
            c_int
        ]
        self.xlib.XFlush.argtypes = [POINTER(Display)]
        self.xlib.XKeysymToKeycode.argtypes = [POINTER(Display), c_ulong]
        self.xlib.XStringToKeysym.argtypes = [c_char_p]
        self.xlib.XSendEvent.argtypes = [
            POINTER(Display), c_ulong, c_int, c_long, c_void_p]
        self.xlib.XQueryTree.argtypes = [
            POINTER(Display),
            c_ulong,
            POINTER(c_ulong),
            POINTER(c_ulong),
            POINTER(POINTER(c_ulong)),
            POINTER(c_uint)
        ]

        # main
        self.display = self.xlib.XOpenDisplay(None)
        self.root_window = self.xlib.XRootWindow(self.display, 0)

    def __getattribute__(self, __name: str):
        if __name in ["xlib", "display", "root_window"]:
            return super().__getattribute__(__name)
        return self.xlib.__getattribute__(__name)


def get_window_property(xlib: Xlib, window_xid: int, property_name: str, return_type: _SimpleCData):
    """
    https://tronche.com/gui/x/xlib/window-information/XGetWindowProperty.html
    """
    actual_type_return = c_ulong()
    actual_format_return = c_int()
    nitems_return = c_ulong()
    bytes_after_return = c_ulong()
    prop_return = POINTER(c_ubyte)()

    xlib.XGetWindowProperty(
        xlib.display,
        window_xid,
        xlib.XInternAtom(
            xlib.display, c_char_p(property_name.encode('utf-8')),
            False
        ),
        0,
        1000,
        False,
        0,  # AnyPropertyType
        byref(actual_type_return),
        byref(actual_format_return),
        byref(nitems_return),
        byref(bytes_after_return),
        byref(prop_return)
    )

    if prop_return:
        data = cast(
            prop_return,
            POINTER(return_type)
        ).contents.value
    else:
        data = None

    # don't forget to free the memory or you will be fucked
    xlib.XFree(prop_return)

    return data


class Window(WindowBase):
    """
    An class for interacting with a window on X11.
    """

    def __init__(self, xid: int, xlib: Xlib) -> None:
        self.xid = xid
        self.xlib = xlib

    @property
    def name(self) -> str:
        name = get_window_property(
            self.xlib,
            self.xid,
            "_NET_WM_NAME",
            c_char * 1024
        )
        if name:
            return name.decode('utf-8')
        return None

    @property
    def pid(self) -> int:
        return get_window_property(self.xlib, self.xid, "_NET_WM_PID", c_long)

    @property
    def active(self) -> bool:
        return self.xid == get_active_window_xid(self.xlib)

    @property
    def geometry(self) -> Box:
        gwa = XWindowAttributes()
        self.xlib.XGetWindowAttributes(self.xlib.display, self.xid, byref(gwa))
        return Box(
            x=gwa.x,
            y=gwa.y,
            width=gwa.width,
            height=gwa.height
        )

    def get_image(self, geometry: Optional[Box] = None) -> Image:
        if geometry is None:
            geometry = self.geometry

        ximage = self.xlib.XGetImage(
            self.xlib.display,  # Display
            self.xid,  # Drawable (Window XID)
            geometry.x - self.geometry.x,  # x
            geometry.y - self.geometry.y,  # y
            geometry.width,  # width
            geometry.height,  # height
            PLAINMASK,  # plane_mask
            ZPIXMAP  # format
        )

        raw_data = cast(
            ximage.contents.data,
            POINTER(c_ubyte * geometry.height * geometry.width * 4)
        )

        data = bytearray(raw_data.contents)

        data = Image(data, geometry.width, geometry.height)

        # don't forget to free the memory or you will be fucked
        self.xlib.XDestroyImage(ximage)

        return data

    def send_chr(self, character: chr) -> None:
        """Send a character to the window

        Args:
            char (str): The character to send
        """
        # https://tronche.com/gui/x/xlib/event-handling/XSendEvent.html

        key = XEvent(type=EventTypes.KeyPress).xkey  # KeyPress
        key.keycode = self.xlib.XKeysymToKeycode(
            self.xlib.display,
            self.xlib.XStringToKeysym(c_char_p(character.encode('utf-8')))
        )  # https://github.com/python-xlib/python-xlib/blob/master/Xlib/keysymdef/latin1.py
        key.window = key.root = self.xid
        key.state = KeyMasks.ShiftMask if character.isupper() else 0

        self.xlib.XSendEvent(
            self.xlib.display,  # Display *display
            key.window,  # Window w
            True,  # Bool propagate
            Masks.KeyPressMask,  # long event_mask
            byref(key)  # XEvent *event_send
        )

        # flush display or events will run delayed cus thai'r only called on the next update
        self.xlib.XFlush(self.xlib.display)

    def send_str(self, string: str) -> None:
        """Send a string to the window

        Args:
            str (str): The string to send
        """
        for character in string:
            self.send_chr(character)

    def warp_pointer(self, x: int, y: int, geometry: Optional[Box] = None) -> None:
        if geometry is None:
            geometry = self.geometry

        # https://tronche.com/gui/x/xlib/input/XWarpPointer.html
        self.xlib.XWarpPointer(
            self.xlib.display,
            self.xid,  # src_w
            self.xid,  # dest_w
            geometry.x,
            geometry.y,
            geometry.width,
            geometry.height,
            x,
            y
        )

        # flush display or events will run delayed cus thai'r only called on the next update
        self.xlib.XFlush(self.xlib.display)

    def send_mouse_click(self, x: int, y: int, button: MouseButtons = MouseButtons.LEFT) -> None:
        # pylint: disable=line-too-long
        """
        Send a mouse click to the window at the given coordinates without moving the pointer.
        Some applications may not respond to the click so it is recommended to also move the pointer with `warp_pointer`.
        """
        # pylint: enable=line-too-long
        event = XEvent(type=EventTypes.ButtonPress).xbutton
        event.window = event.root = self.xid
        event.button = button
        event.x_root = 0
        event.y_root = 0
        event.x = x
        event.y = y

        self.xlib.XSendEvent(
            self.xlib.display,
            event.window,
            True,
            Masks.ButtonPressMask,
            byref(event)
        )

        # flush display or events will run delayed cus thai'r only called on the next update
        self.xlib.XFlush(self.xlib.display)

        event.type = EventTypes.ButtonRelease

        self.xlib.XSendEvent(
            self.xlib.display,
            event.window,
            True,
            Masks.ButtonReleaseMask,
            byref(event)
        )

        # flush display or events will run delayed cus thai'r only called on the next update
        self.xlib.XFlush(self.xlib.display)


def get_active_window_xid(xlib: Xlib) -> int:
    """
    Returns the XID of the active window.
    """
    return get_window_property(
        xlib,
        xlib.root_window,
        "_NET_ACTIVE_WINDOW",
        c_long
    )


def get_connected_xids(xlib: Xlib, window: int):
    """
    https://tronche.com/gui/x/xlib/window-information/XQueryTree.html\n
    Uses XQueryTree to get the XIDs of connected windows.
    """
    root_return = c_ulong()
    parent_return = c_ulong()
    children_return = POINTER(c_ulong)()
    nitems_return = c_uint()

    xlib.XQueryTree(
        xlib.display,  # Display *display
        window,  # Window window
        byref(root_return),  # Window *root_return
        byref(parent_return),  # Window *parent_return
        byref(children_return),  # Window **children_return
        byref(nitems_return)  # unsigned int *nitems_return
    )

    xids = []

    for xid in range(nitems_return.value):
        xids.append(children_return[xid])

    # don't forget to free the memory or you will be fucked
    xlib.XFree(children_return)

    return xids


def get_all_windows(xlib: Xlib) -> list:
    """
    Get all window XIDs. By recursively getting all connected windows.
    """
    final = get_connected_xids(xlib, xlib.root_window)
    next_window = final.copy()

    run = True
    while run:
        run = False
        next_temp = []
        for xid in next_window:
            xids = get_connected_xids(xlib, xid)
            if len(xids) > 0:
                run = True
            next_temp += xids

        next_window = next_temp
        final += next_window

    final_windows = []

    for xid in final:
        final_windows.append(Window(xid, xlib))

    return final_windows


class DSI(DSIBase):
    """
    Main DSI class
    """

    def __init__(self):
        self.xlib = Xlib()

    def get_active_window(self) -> WindowBase:
        return Window(get_active_window_xid(self.xlib), self.xlib)

    def get_all_windows(self) -> list:
        return get_all_windows(self.xlib)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module provides a function to detect the current Linux display server.
"""

from os import getenv
from enum import Enum


class LinuxDisplayServer(Enum):
    """
    An enumeration of Linux display servers.
    """
    WAYLAND = "wayland"
    X11 = "x11"
    UNKNOWN = "Unknown"


def detect_linux_display_server() -> LinuxDisplayServer:
    """
    Detects the current Linux display server.
    """
    xdg_session_type = getenv("XDG_SESSION_TYPE")
    if xdg_session_type == "wayland":
        return LinuxDisplayServer.WAYLAND

    if xdg_session_type == "x11":
        return LinuxDisplayServer.X11

    if getenv("WAYLAND_DISPLAY") is not None:
        return LinuxDisplayServer.WAYLAND

    if getenv("DISPLAY") is not None:
        return LinuxDisplayServer.X11

    # maby check if a process is called "Xwayland" or "Xorg"

    return LinuxDisplayServer.UNKNOWN



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""Exceptions for DSI."""


class DisplayServerNotSupportedError(Exception):
    """Exception raised when the display server is not supported."""


class OSNotSupportedError(Exception):
    """Exception raised when the operating system is not supported."""



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module provides a class for interacting with windows
"""

# built-in modules
from abc import ABCMeta, abstractmethod
from typing import Optional

# local modules
from .image import Image
from .buttons import MouseButtons
from .box import Box


class WindowBase(metaclass=ABCMeta):
    """
    An abstract base class that defines the interface for interacting with a window.
    """
    @property
    @abstractmethod
    def name(self) -> str:
        """
        Returns the name of the window.
        Returns None if no name is available.
        """

    @property
    @abstractmethod
    def pid(self) -> int:
        """
        Returns the process id of the window.
        Returns None if no pid is available.
        """

    @property
    @abstractmethod
    def active(self) -> bool:
        """
        Returns True if the window is active.
        Returns False if the window is not active.
        """

    @property
    @abstractmethod
    def geometry(self) -> Box:
        """
        Returns: tuple: (x, y, width, height)
        """

    @abstractmethod
    def get_image(self, geometry: Optional[Box] = None) -> Image:
        # pylint: disable=line-too-long
        """
        Returns an Image of the window.
        With the geometry parameter you can specify a sub-region of the window that will be captured.
        """
        # pylint: enable=line-too-long

    @abstractmethod
    def send_chr(self, character: chr) -> None:
        """
        send the keystroke of the given character to the window.
        """

    @abstractmethod
    def send_str(self, string: str) -> None:
        """
        Send keystrokes equivalent to the string you pass to the window.
        """

    @abstractmethod
    # pylint: disable-next=invalid-name
    def warp_pointer(self, x: int, y: int, geometry: Optional[Box] = None) -> None:
        """
        Moves the pointer relative to the window to the given coordinates.
        """

    @abstractmethod
    # pylint: disable-next=invalid-name
    def send_mouse_click(self, x: int, y: int, button: MouseButtons = MouseButtons.LEFT) -> None:
        """
        Send a mouse click to the window at the given coordinates.
        On some windows/applications you need to move the pointer with warp_pointer() first.
        """

    def __repr__(self) -> str:
        # pylint: disable=line-too-long
        name = self.name
        if name:
            return f'Window(name="{self.name}", pid={self.pid}, active={self.active}, geometry={self.geometry})'
        return f'Window(pid={self.pid}, active={self.active}, geometry={self.geometry})'
        # pylint: enable=line-too-long



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module provides a Box class that represents a rectangle with a position and size.
"""


class Box(tuple):
    """
    A Box represents a rectangle with a position and size.
    """

    def __init__(self, x: int, y: int, width: int, height: int):
        """This is just here for autocompletion"""

    def __new__(cls, x: int, y: int, width: int, height: int):
        return tuple.__new__(Box, (x, y, width, height))

    @property
    # pylint: disable-next=invalid-name
    def x(self) -> int:
        """
        Gets the x-coordinate of the box.
        """
        return self[0]

    @property
    # pylint: disable-next=invalid-name
    def y(self) -> int:
        """
        Gets the y-coordinate of the box.
        """
        return self[1]

    @property
    def width(self) -> int:
        """
        Gets the width of the box.
        """
        return self[2]

    @property
    def height(self) -> int:
        """
        Gets the height of the box.
        """
        return self[3]

    def __repr__(self) -> str:
        return f'Box(x={self.x}, y={self.y}, width={self.width}, height={self.height})'


def main() -> None:
    """
    Creates and displays an example Box object.
    """
    try:
        # pylint: disable-next=import-outside-toplevel, redefined-builtin
        from rich import print
    except ImportError:
        pass
    box = Box(100, 200, 300, 400)
    print(box)
    print(f"x={box.x}")
    print(f"y={box.y}")
    print(f"width={box.width}")
    print(f"height={box.height}")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
All Windows specific DSI functions
"""

# built-in modules
from typing import Optional
from ctypes import (
    windll,
    WINFUNCTYPE,
    c_long,
    Structure,
    create_string_buffer
)
from ctypes.wintypes import (
    DWORD,
    BOOL,
    HWND,
    LPARAM
)
from ctypes import (
    byref,
    create_unicode_buffer,
    sizeof
)

# local modules
from .base import DSIBase
from .window import WindowBase
from .image import Image
from .buttons import MouseButtons
from .box import Box

user32 = windll.user32
gdi32 = windll.gdi32

"""
https://learn.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-bitblt#SRCCOPY
Copies the source rectangle directly to the destination rectangle.
"""
SRCCOPY = 0x00CC0020
WM_CHAR = 0x0102
WM_KEYDOWN = 0x0100
WM_KEYUP = 0x0101
# pylint: disable=too-few-public-methods

# https://learn.microsoft.com/de-de/windows/win32/inputdev/virtual-key-codes
vk_map = {
    "lbutton": 0x01,  # VK_LBUTTON
    "rbutton": 0x02,  # VK_RBUTTON
    "cancel": 0x03,  # VK_CANCEL
    "mbutton": 0x04,  # VK_MBUTTON
    "xbutton1": 0x05,  # VK_XBUTTON1
    "xbutton2": 0x06,  # VK_XBUTTON2
    # - 0x07
    "back": 0x08,  # VK_BACK
    "tab": 0x09,  # VK_TAB
    # - 0x0A-0B
    "clear": 0x0C,  # VK_CLEAR
    "return": 0x0D,  # VK_RETURN
    # - 0x0E-0F
    "shift": 0x10,  # VK_SHIFT
    "control": 0x11,  # VK_CONTROL (STRG)
    "menu": 0x12,  # VK_MENU (ALT)
    "pause": 0x13,  # VK_PAUSE
    "capital": 0x14,  # VK_CAPITAL (caps lock)
    "kana": 0x15,  # VK_KANA
    "hangul": 0x15,  # VK_HANGUEL
    "ime_on": 0x16,  # VK_IME_ON
    "junja": 0x17,  # VK_JUNJA
    "final": 0x18,  # VK_FINAL
    "hanja": 0x19,  # VK_HANJA
    "kanji": 0x19,  # VK_KANJI
    "ime_off": 0x1A,  # VK_IME_OFF
    "escape": 0x1B,  # VK_ESCAPE
    "convert": 0x1C,  # VK_CONVERT
    "nonconvert": 0x1D,  # VK_NONCONVERT
    "accept": 0x1E,  # VK_ACCEPT
    "modechange": 0x1F,  # VK_MODECHANGE
    "space": 0x20,  # VK_SPACE
    "prior": 0x21,  # VK_PRIOR
    "next": 0x22,  # VK_NEXT
    "end": 0x23,  # VK_END
    "home": 0x24,  # VK_HOME
    "left": 0x25,  # VK_LEFT
    "up": 0x26,  # VK_UP
    "right": 0x27,  # VK_RIGHT
    "down": 0x28,  # VK_DOWN
    "select": 0x29,  # VK_SELECT
    "print": 0x2A,  # VK_PRINT
    "execute": 0x2B,  # VK_EXECUTE
    "snapshot": 0x2C,  # VK_SNAPSHOT
    "insert": 0x2D,  # VK_INSERT
    "delete": 0x2E,  # VK_DELETE
    "help": 0x2F,  # VK_HELP
    # 0-9 0x30-0x39
    # - 0x3A-40
    # A-Z 0x41-0x5A
    "lwin": 0x5B,  # VK_LWIN
    "rwin": 0x5C,  # VK_RWIN
    "apps": 0x5D,  # VK_APPS
    # - 0x5E
    "sleep": 0x5F,  # VK_SLEEP
    "numpad0": 0x60,  # VK_NUMPAD0
    "numpad1": 0x61,  # VK_NUMPAD1
    "numpad2": 0x62,  # VK_NUMPAD2
    "numpad3": 0x63,  # VK_NUMPAD3
    "numpad4": 0x64,  # VK_NUMPAD4
    "numpad5": 0x65,  # VK_NUMPAD5
    "numpad6": 0x66,  # VK_NUMPAD6
    "numpad7": 0x67,  # VK_NUMPAD7
    "numpad8": 0x68,  # VK_NUMPAD8
    "numpad9": 0x69,  # VK_NUMPAD9
    "multiply": 0x6A,  # VK_MULTIPLY
    "add": 0x6B,  # VK_ADD
    "separator": 0x6C,  # VK_SEPARATOR
    "subtract": 0x6D,  # VK_SUBTRACT
    "decimal": 0x6E,  # VK_DECIMAL
    "divide": 0x6F,  # VK_DIVIDE
    "f1": 0x70,  # VK_F1
    "f2": 0x71,  # VK_F2
    "f3": 0x72,  # VK_F3
    "f4": 0x73,  # VK_F4
    "f5": 0x74,  # VK_F5
    "f6": 0x75,  # VK_F6
    "f7": 0x76,  # VK_F7
    "f8": 0x77,  # VK_F8
    "f9": 0x78,  # VK_F9
    "f10": 0x79,  # VK_F10
    "f11": 0x7A,  # VK_F11
    "f12": 0x7B,  # VK_F12
    "f13": 0x7C,  # VK_F13
    "f14": 0x7D,  # VK_F14
    "f15": 0x7E,  # VK_F15
    "f16": 0x7F,  # VK_F16
    "f17": 0x80,  # VK_F17
    "f18": 0x81,  # VK_F18
    "f19": 0x82,  # VK_F19
    "f20": 0x83,  # VK_F20
    "f21": 0x84,  # VK_F21
    "f22": 0x85,  # VK_F22
    "f23": 0x86,  # VK_F23
    "f24": 0x87,  # VK_F24
    # - 0x88-8F
    "numlock": 0x90,  # VK_NUMLOCK
    "scroll": 0x91,  # VK_SCROLL
    # OEM 0x92-96
    # - 0x97-9F
    "lshift": 0xA0,  # VK_LSHIFT
    "rshift": 0xA1,  # VK_RSHIFT
    "lcontrol": 0xA2,  # VK_LCONTROL
    "rcontrol": 0xA3,  # VK_RCONTROL
    "lmenu": 0xA4,  # VK_LMENU
    "rmenu": 0xA5,  # VK_RMENU
    "browser_back": 0xA6,  # VK_BROWSER_BACK
    "browser_forward": 0xA7,  # VK_BROWSER_FORWARD
    "browser_refresh": 0xA8,  # VK_BROWSER_REFRESH
    "browser_stop": 0xA9,  # VK_BROWSER_STOP
    "browser_search": 0xAA,  # VK_BROWSER_SEARCH
    "browser_favorites": 0xAB,  # VK_BROWSER_FAVORITES
    "browser_home": 0xAC,  # VK_BROWSER_HOME
    "volume_mute": 0xAD,  # VK_VOLUME_MUTE
    "volume_down": 0xAE,  # VK_VOLUME_DOWN
    "volume_up": 0xAF,  # VK_VOLUME_UP
    "media_next_track": 0xB0,  # VK_MEDIA_NEXT_TRACK
    "media_prev_track": 0xB1,  # VK_MEDIA_PREV_TRACK
    "media_stop": 0xB2,  # VK_MEDIA_STOP
    "media_play_pause": 0xB3,  # VK_MEDIA_PLAY_PAUSE
    "launch_mail": 0xB4,  # VK_LAUNCH_MAIL
    "launch_media_select": 0xB5,  # VK_LAUNCH_MEDIA_SELECT
    "launch_app1": 0xB6,  # VK_LAUNCH_APP1
    "launch_app2": 0xB7,  # VK_LAUNCH_APP2
    # - 0xB8-B9
    "oem_1": 0xBA,  # VK_OEM_1
    "oem_plus": 0xBB,  # VK_OEM_PLUS
    "oem_comma": 0xBC,  # VK_OEM_COMMA
    "oem_minus": 0xBD,  # VK_OEM_MINUS
    "oem_period": 0xBE,  # VK_OEM_PERIOD
    "oem_2": 0xBF,  # VK_OEM_2
    "oem_3": 0xC0,  # VK_OEM_3
    # - 0xC1-D7
    # - 0xD8-DA
    "oem_4": 0xDB,  # VK_OEM_4
    "oem_5": 0xDC,  # VK_OEM_5
    "oem_6": 0xDD,  # VK_OEM_6
    "oem_7": 0xDE,  # VK_OEM_7
    "oem_8": 0xDF,  # VK_OEM_8
    # - 0xE0
    # OEM 0xE1
    "oem_102": 0xE2,  # VK_OEM_102
    # OEM 0xE3-E4
    "processkey": 0xE5,  # VK_PROCESSKEY
    # OEM 0xE6
    "packet": 0xE7,  # VK_PACKET
    # - 0xE8
    # OEM 0xE9-F5
    "attn": 0xF6,  # VK_ATTN
    "crsel": 0xF7,  # VK_CRSEL
    "exsel": 0xF8,  # VK_EXSEL
    "ereof": 0xF9,  # VK_EREOF
    "play": 0xFA,  # VK_PLAY
    "zoom": 0xFB,  # VK_ZOOM
    "noname": 0xFC,  # VK_NONAME
    "pa1": 0xFD,  # VK_PA1
    "oem_clear": 0xFE,  # VK_OEM_CLEAR
}


class RECT(Structure):
    """
    A structure that defines the coordinates of a rectangle.
    """
    _fields_ = [
        ("left", c_long),
        ("top", c_long),
        ("right", c_long),
        ("bottom", c_long)
    ]

# pylint: enable=too-few-public-methods


class Window(WindowBase):
    """
    An class for interacting with a window on Windows.
    """

    def __init__(self, window) -> None:
        self.window = window

    @property
    def name(self) -> str:
        buffer = create_unicode_buffer(1024)
        user32.GetWindowTextW(self.window, buffer, sizeof(buffer))
        return buffer.value

    @property
    def pid(self) -> int:
        pid = DWORD()
        user32.GetWindowThreadProcessId(self.window, byref(pid))
        return pid.value

    @property
    def active(self) -> bool:
        return self.window == user32.GetForegroundWindow()

    @property
    def geometry(self) -> Box:
        rect = RECT()
        user32.GetClientRect(self.window, byref(rect))

        window_rect = RECT()
        user32.GetWindowRect(self.window, byref(window_rect))

        title_bar_height = window_rect.bottom - \
            window_rect.top - (rect.bottom - rect.top)

        return Box(
            x=window_rect.left,
            y=window_rect.top + title_bar_height,
            width=rect.right - rect.left,
            height=rect.bottom - rect.top
        )

    def get_image(self, geometry: Optional[Box] = None):
        if geometry is None:
            geometry = self.geometry

        wdc = user32.GetDC(self.window)

        # adjust the bitmap size
        data_bitmap = gdi32.CreateCompatibleBitmap(
            wdc,
            geometry.width,
            geometry.height
        )

        # get the window image data
        dc_object = gdi32.CreateCompatibleDC(None)
        gdi32.SelectObject(dc_object, data_bitmap)
        gdi32.BitBlt(
            dc_object,
            0,
            0,
            geometry.width,
            geometry.height,
            wdc,
            geometry.x - self.geometry.x,
            geometry.y - self.geometry.y,
            SRCCOPY
        )

        # convert the raw data into a format opencv can read
        signed_ints_array = create_string_buffer(
            geometry.width * geometry.height * 4
        )
        gdi32.GetBitmapBits(
            data_bitmap,
            geometry.width * geometry.height * 4,
            signed_ints_array
        )
        img = Image(signed_ints_array, geometry.width, geometry.height)

        # free resources
        gdi32.DeleteObject(data_bitmap)
        gdi32.DeleteDC(dc_object)
        user32.ReleaseDC(self.window, wdc)

        return img

    def send_chr(self, character: chr) -> None:
        vk = vk_map.get(character.lower())
        if vk is None:
            user32.PostMessageW(self.window, WM_CHAR, ord(character), 0)
        else:
            user32.PostMessageW(self.window, WM_KEYDOWN, vk, 0)
            user32.PostMessageW(self.window, WM_KEYUP, vk, 0)

    def send_str(self, string: str) -> None:
        for character in string:
            self.send_chr(character)

    def warp_pointer(self, x: int, y: int, geometry: Optional[Box] = None) -> None:
        if geometry is None:
            geometry = self.geometry
        user32.SetCursorPos(x + geometry.x, y + geometry.y)

    def send_mouse_click(self, x: int, y: int, button: MouseButtons = MouseButtons.LEFT) -> None:
        # Send the mouse click event
        if button == MouseButtons.LEFT:
            # Left button down
            user32.PostMessageW(
                self.window,
                0x201,
                0,
                x | (y << 16)
            )
            # Left button up
            user32.PostMessageW(
                self.window,
                0x202,
                0,
                x | (y << 16)
            )
        elif button == MouseButtons.RIGHT:
            # Right button down
            user32.PostMessageW(
                self.window,
                0x204,
                0,
                x | (y << 16)
            )
            # Right button up
            user32.PostMessageW(
                self.window,
                0x205,
                0,
                x | (y << 16)
            )
        elif button == MouseButtons.MIDDLE:
            # Middle button down
            user32.PostMessageW(
                self.window,
                0x207,
                0,
                x | (y << 16)
            )
            # Middle button up
            user32.PostMessageW(
                self.window,
                0x208,
                0,
                x | (y << 16)
            )
        elif button == MouseButtons.BACKWARD:
            # Backward button down
            user32.PostMessageW(
                self.window,
                0x020D,
                0,
                x | (y << 16)
            )
            # Backward button up
            user32.PostMessageW(
                self.window,
                0x020E,
                0,
                x | (y << 16)
            )
        elif button == MouseButtons.FORWARD:
            # Forward button down
            user32.PostMessageW(
                self.window,
                0x020B,
                0,
                x | (y << 16)
            )
            # Forward button up
            user32.PostMessageW(
                self.window,
                0x020C,
                0,
                x | (y << 16)
            )
        else:
            raise ValueError(f"Invalid button code '{button}'.")


class DSI(DSIBase):
    """
    Main DSI class
    """

    def __init__(self):
        pass

    def get_active_window(self) -> WindowBase:
        # GetDesktopWindow
        return Window(user32.GetForegroundWindow())

    def get_all_windows(self) -> list[WindowBase]:
        windows = []

        @WINFUNCTYPE(BOOL, HWND, LPARAM)
        def callback(hwnd, _unused):
            windows.append(Window(hwnd))
            return True

        user32.EnumWindows(callback)

        return windows



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This is an example of DSI
"""

# local modules
from . import DSI


def main() -> None:
    """
    This is an example of DSI
    """
    with DSI() as dsi:
        window = dsi.get_active_window()

        print("Active window: ")
        print(f"\tName: {window.name}")
        print(f"\tPID: {window.pid}")
        if window.xid:
            print(f"\tXID: {window.xid}")
        print(f"\tGeometry: {window.geometry}")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
The Image module contains a class to hold the raw data of an image.
"""

from typing import Tuple


class Image:
    """
    A class to that holds the raw data of an image.
    Use np.array(Image) to get a numpy array of the image.
    """

    def __init__(self, data, width, height):
        self.data = data
        self.width = width
        self.height = height

    @property
    def __array_interface__(self) -> dict:
        """
        A property method that returns an interface
        for the numpy array interface.
        https://docs.scipy.org/doc/numpy/reference/arrays.interface.html
        """
        return {
            "version": 3,
            "shape": (self.height, self.width, 4),
            "typestr": "|u1",
            "data": self.data,
        }

    def __repr__(self) -> str:
        return f"Image(width={self.width}, height={self.height})"

    # pylint: disable-next=invalid-name
    def get_pixel(self, x: int, y: int) -> Tuple[int, int, int, int]:
        """
        This function retrieves the RGBA values of the pixel,
        located at the specified coordinates
        and returns them as a 4-tuple integer.
        """
        pixel_start_index = (y * self.width + x) * 4
        pixel_data = self.data[pixel_start_index:pixel_start_index + 4]
        return tuple(pixel_data)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module initializes DSI
"""

# built-in modules
from platform import system as __system

# local modules
from .exceptions import (
    DisplayServerNotSupportedError as __DisplayServerNotSupportedError,
    OSNotSupportedError as __OSNotSupportedError
)

__version__ = "0.0.dev11"
__author__ = "Commandcracker"

__os_name = __system().lower()

if __os_name == "linux":
    from .util import detect_linux_display_server as __detect_linux_display_server
    from .util import LinuxDisplayServer as __LinuxDisplayServer
    __linux_display_server = __detect_linux_display_server()

    if __linux_display_server is __LinuxDisplayServer.X11:
        from .linux import DSI
    elif __linux_display_server is __LinuxDisplayServer.WAYLAND:
        raise NotImplementedError("Wayland is not yet implemented.")
    else:
        raise __DisplayServerNotSupportedError(
            "Your display server is not supported."
        )

elif __os_name == "windows":
    from .windows import DSI

elif __os_name == "darwin":
    raise NotImplementedError("MacOS is not yet implemented.")

else:
    raise __OSNotSupportedError("Your OS is not supported.")

__all__ = ["DSI"]



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module holds the main DSI class
"""


# built-in modules
from abc import ABCMeta, abstractmethod
from platform import system

# local modules
from .window import WindowBase


class DSIBase(metaclass=ABCMeta):
    """
    Main DSI class
    """
    @abstractmethod
    def get_active_window(self) -> WindowBase:
        """
        Gets the active window.
        Returns None if no window is active.
        """

    @abstractmethod
    def get_all_windows(self) -> list[WindowBase]:
        """
        Returns a list of all Windows.
        """

    def get_window_by_pid(self, pid: int) -> WindowBase:
        """
        Get window by pid.
        Returns None if no window found.
        """
        all_window = self.get_all_windows()

        for window in all_window:
            if window.pid == pid:
                return window
        return None

    def get_window_by_name(self, name: str) -> WindowBase:
        """
        Get a window by name.
        Returns None if no window with that name is found.
        """
        all_window = self.get_all_windows()

        for window in all_window:
            if window.name is not None and name in window.name:
                return window
        return None

    @property
    def platform(self) -> str:
        """
        Returns the platform name.
        """
        return system().lower()

    @property
    def linux(self) -> bool:
        """
        Returns True if the platform is linux.
        """
        return self.platform == "linux"

    @property
    def windows(self) -> bool:
        """
        Returns True if the platform is windows.
        """
        return self.platform == "windows"

    @property
    def mac(self) -> bool:
        """
        Returns True if the platform is mac.
        """
        return self.platform == "darwin"

    # Allow dsi to be called with ’with’

    def __enter__(self):
        return self

    def __exit__(self, *_):
        pass



from setuptools import setup, find_packages
from os.path import dirname, join, abspath
# def read_requirements(file):
#    with open(file) as f:
#        return f.read().splitlines()


def read(rel_path: str) -> str:
    here = abspath(dirname(__file__))
    with open(join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


# requirements = read_requirements("requirements.txt")

def main() -> None:
    setup(
        name="display-server-interactions",
        version=get_version("display_server_interactions/__init__.py"),

        author="Commandcracker",

        # url="https://github.com/Commandcracker/display-server-interactions",

        project_urls={
            "Documentation": "https://display-server-interactions.readthedocs.io/en/latest/",
            "Source": "https://github.com/Commandcracker/display-server-interactions",
            "Tracker": "https://github.com/Commandcracker/display-server-interactions/issues"
        },

        description="DSI allows you to perform basic interactions on your display server, like screenshotting a window or sending input to it.",
        long_description_content_type="text/markdown",
        long_description=read("README.md"),

        py_modules=["display_server_interactions"],

        keywords=[
            "screen",
            "display",
            "screenshot",
            "input",
            "window",
            "screencapture",
            "screengrab",
        ],

        license="Apache Software License",
        packages=find_packages(exclude=["test"]),
        # install_requires=requirements,
        classifiers=[
            "Development Status :: 2 - Pre-Alpha",

            "License :: OSI Approved :: Apache Software License",

            "Operating System :: POSIX :: Linux",
            "Operating System :: Microsoft :: Windows",
            # "Operating System :: OS Independent",

            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",

            "Topic :: Multimedia :: Graphics",
            "Topic :: Multimedia :: Graphics :: Capture :: Screen Capture",
        ]
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module contains everything related too buttons
"""

# pylint: disable=too-few-public-methods


class MouseButtons:
    """
    https://tronche.com/gui/x/xlib/events/keyboard-pointer/keyboard-pointer.html\n
    /usr/include/X11/X.h: 259-263
    """
    LEFT = 1
    RIGHT = 2
    MIDDLE = 3
    FORWARD = 4
    BACKWARD = 5

# pylint: enable=too-few-public-methods
